from FrozenYoghourt import *
    

class Mode:
    """
    A class to help with switching between Numpy and Sympy
    
    Attributes
    ----------
    representation: str
        a string to represent the current mode ('numpy' or 'sympy')
        
    Methods
    -------
    toggle(specific_mode = None, printing = True)
        Switch to an alternate mode, or to a specific inputted mode
    now()
        Print out the current mode
    
    """
    
    
    
    representation = 'numpy'
    
    @classmethod
    def toggle(cls, specific_mode = None, printing = True):
        """
        Switch to an alternate mode, or to a specific inputted mode
        
        If the argument specific_mode is not passed in, the default None is used
        which switch to the alternate mode
        
        Parameters
        ----------
        specific_mode: str, optional
            Mode to switch to (default is None)
        printing: bool, optional
            Option to print a message when switching mode (default is True)
        
        """

        if specific_mode is None:
            cls.representation = {'numpy': 'sympy',
                                  'sympy': 'numpy'}[cls.representation]
            if printing:
                print(f'\033[1;34m{cls.representation.upper()} \033[1;32mactivated')
        else:
            cls.representation = {'n': 'numpy', 
                                  'numpy': 'numpy', 
                                  's': 'sympy', 
                                  'sympy':'sympy'}[specific_mode]
            if printing:
                print(f'\033[1;34m{cls.representation.upper()} \033[1;32mactivated')
            
    @classmethod
    def now(cls):
        """
        Print out the current mode
        """
        
        print(f'\033[1;34m{cls.representation.upper()} \033[1;30mis\033[1;30m \033[1;32mon')


from FrozenYoghourt import *
from FrozenYoghourt.mode import *
from FrozenYoghourt.maths import *
from FrozenYoghourt.gates import *

def local_ops(num_qubits = 1, no_ops = 1, unimodular = False):
    
    if type(no_ops) == bool:
        unimodular = no_ops; no_ops = 1
    
    if no_ops == 1:
        return tp(*[u2(unimodular, i) for i in range(num_qubits)])
    else:
        return [tp(*[u2(unimodular, i) for i in range(j*num_qubits, (j+1)*num_qubits)]) for j in range(no_ops)]

def epsilon(psi, num_qubits=2):
    if Mode.representation == 'numpy':
        return np.abs(psi.T @ tp(Y(), no_times=num_qubits) @ psi)[0]
    else:
        return Abs(psi.T @ tp(Y(), no_times=num_qubits) @ psi)[0]

def global_phase(A, B):
    D = np.diag(np.conj(A).T @ B)

    if np.isclose(np.min(D), np.max(D)):
        return D[0]
    else:
        print('A and B are not equivalent up to global phase')
        return False

def kron_decomp(U:np.ndarray):
    m = U.reshape(2, 2, 2, 2).transpose(0, 2, 1, 3).reshape(4, 4)

    u, sv, vh = np.linalg.svd(m)

    a = np.sqrt(sv[0]) * u[:, 0].reshape(2, 2)
    b = np.sqrt(sv[0]) * vh[0, :].reshape(2, 2)

    return a, b


def gamma(U, unimodular = False):
    n = int(np.log2(U.shape[0]))

    if (Mode.repsentation == 'numpy') and (not np.isclose(np.linalg.det(U), 1)):
        U = to_su(U)
    elif unimodular:
        U = to_su(U)
        
    E = tp(Y(), no_times=n)
    return U @ E @ U.T @ E

def chi(M, variable=None):
    dim = M.shape[0]

    coef = np.array([1])
    Mk = np.array(M)

    for k in range(1, dim + 1):
        ak = -Mk.trace() / k
        coef = np.append(coef, ak)
        Mk += np.diag(np.repeat(ak, dim))
        Mk = np.dot(M, Mk)

    if Mode.representation == 'numpy':
        return coef

    else:
        if variable is None:
            return Matrix(coef)
        else:
            var = symbols(variable)
            variable_mat = Matrix([var ** n for n in range(dim + 1)])
            coef = Matrix(coef).T
            char_poly = simplify(coef @ variable_mat)[0]
            return char_poly

def shende_invariant(U, V = None, return_charpoly = False):

    if V is None:
        return chi(gamma(to_su(U)))
    
    elif Mode.representation == 'numpy':
        charpoly_U = chi(gamma(to_su(U)))
        charpoly_V = chi(gamma(to_su(V)))
        
        if return_charpoly:
            return np.all(np.isclose(charpoly_U, charpoly_V)), charpoly_U, charpoly_V
        else:
            return np.all(np.isclose(charpoly_U, charpoly_V))
        
    else: ## This needs fixing because the determinant function doesn't work
        charpoly_U = chi(gamma(to_su(U)))
        charpoly_V = chi(gamma(to_su(V)))
        
        if return_charpoly:
            return charpoly_U.equals(charpoly_V), charpoly_U, charpoly_V
        else:
            return charpoly_U.equals(charpoly_V)


def canonical_class_vector(x, y, z, return_phase = True):
    
    # Step 1 and 2
    param = np.array([x, y, z])
    shift = param // np.pi
    phase = np.exp(-1j*(np.pi/2)*(np.sum(shift)%2)) # Keep track of whether to add phase to correct for KAK
    k = np.sort(param - np.pi*shift)[::-1]

    # Step 3
    if k[0] + k[1] > np.pi:
        k = np.sort(np.array([np.pi - k[1], np.pi - k[0], k[2]]))[::-1]

    # Step 4
    if np.isclose(k[2], 0) and (k[0] > np.pi/2):
        k[0] = np.pi - k[0]
        k = np.sort(k)[::-1]
        phase *= 1j
    
    if return_phase:
        return k[0], k[1], k[2], phase
    else:
        return k[0], k[1], k[2]

### Add huang_invariant tomorrow


from FrozenYoghourt import *
from FrozenYoghourt.mode import *
from FrozenYoghourt.maths import *



def Id():
    if Mode.representation == 'numpy':
        return np.array([[1, 0], [0, 1]])
    else:
        return Matrix([[1, 0], [0, 1]])

def X():
    if Mode.representation == 'numpy':
        return np.array([[0, 1], [1, 0]])
    else:
        return Matrix([[0, 1], [1, 0]])

def Y():
    if Mode.representation == 'numpy':
        return np.array([[0, -1j], [1j, 0]])
    else:
        return Matrix([[0, -I], [I, 0]])

def Z():
    if Mode.representation == 'numpy':
        return np.array([[1, 0], [0, -1]])
    else:
        return Matrix([[1, 0], [0, -1]])

def H():
    if Mode.representation == 'numpy':
        return 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
    else:
        return 1 / sqrt(2) * Matrix([[1, 1], [1, -1]])

def Rx(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.cos(theta / 2), -1j * np.sin(theta / 2)],
                         [-1j * np.sin(theta / 2), np.cos(theta / 2)]])
    else:
        return Matrix([[cos(theta / 2), -I * sin(theta / 2)],
                       [-I * sin(theta / 2), cos(theta / 2)]])

def Ry(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.cos(theta / 2), -np.sin(theta / 2)],
                         [np.sin(theta / 2), np.cos(theta / 2)]])
    else:
        return Matrix([[cos(theta / 2), -sin(theta / 2)],
                       [sin(theta / 2), cos(theta / 2)]])

def Rz(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.exp(-1j * theta / 2), 0],
                         [0, np.exp(1j * theta / 2)]])
    else:
        return Matrix([[exp(-I * theta / 2), 0],
                       [0, exp(I * theta / 2)]])

def Phase(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[1, 0], [0, np.exp(1j*theta)]])
    else:
        return Matrix([[1, 0], [0, exp(I*theta)]])

def u2(angles:list = None, unimodular = False, index = ''):
    
    if type(angles) == bool:
        unimodular = angles
        angles = None
    
    if Mode.representation == 'numpy':
        if unimodular: 
            return to_su(random_unitary(2).data)
        else: 
            return random_unitary(2).data
    else:
        if angles is None:
            theta, phi, lam = symbols(f'theta_{index}, phi_{index}, lambda_{index}', real = True)
        else:
            theta, phi, lam = angles
        if unimodular:
            return Matrix([[exp(I*phi)*cos(theta / 2), -exp(I * -lam) * sin(theta / 2)],
                                          [exp(I * lam) * sin(theta / 2), exp(I * (-phi)) * cos(theta / 2)]])
        else:
            return Matrix([[cos(theta / 2), -exp(I * lam) * sin(theta / 2)],
                                           [exp(I * phi) * sin(theta / 2), exp(I * (phi + lam)) * cos(theta / 2)]])

def CX():
    if Mode.representation == 'numpy':
        return np.array([[1, 0, 0, 0],
                         [0, 1, 0, 0],
                         [0, 0, 0, 1],
                         [0, 0, 1, 0]])
    else:
        return Matrix([[1, 0, 0, 0],
                       [0, 1, 0, 0],
                       [0, 0, 0, 1],
                       [0, 0, 1, 0]])

def CU(control, target, U, no_qubits = 2):
    """
    Manually build the unitary matrix for non-adjacent CU gates
    Parameters:
    -----------
    control: int
        Index of the control qubit (1st qubit is index 0)
    target: int
        Index of the target qubit (1st qubit is index 0)
    U: ndarray
        Target unitary matrix
    edian: bool (True: qiskit convention)
        Qubits order convention
    no_qubits: int
        Number of qubits in the circuit
    Returns:
    --------
    cx_out:
        Unitary matrix for CU gate
    """

    left = [Id()] * no_qubits
    right = [Id()] * no_qubits

    if Mode.representation == 'numpy':

        left[control] = np.array([[1, 0], [0, 0]])
        right[control] = np.array([[0, 0], [0, 1]])

    else:

        left[control] = Matrix([[1, 0], [0, 0]])
        right[control] = Matrix([[0, 0], [0, 1]])

    right[target] = U

    cu_out = tp(*left) + tp(*right)

    return cu_out

def Swap():
    if Mode.representation == 'numpy':
        return np.array([[1, 0, 0, 0],
                         [0, 0, 1, 0],
                         [0, 1, 0, 0],
                         [0, 0, 0, 1]])
    else:
        return Matrix([[1, 0, 0, 0],
                       [0, 0, 1, 0],
                       [0, 1, 0, 0],
                       [0, 0, 0, 1]])

def SQiSW():
    if Mode.representation == 'numpy':
        return np.array([[1, 0, 0, 0],
                         [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
                         [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0],
                         [0, 0, 0, 1]])
    else:
        return Matrix([[1, 0, 0, 0],
                       [0, 1 / sqrt(2), I / sqrt(2), 0],
                       [0, I / sqrt(2), 1 / sqrt(2), 0],
                       [0, 0, 0, 1]])

def Gamma():
    if Mode.representation == 'numpy':
        return 1/2 * np.array([[1, -1, 1, -1], 
                               [1, 1, -1, -1], 
                               [1, -1, -1, 1], 
                               [1, 1, 1, 1]])
    else:
        return 1/2 * Matrix([[1, -1, 1, -1], 
                             [1, 1, -1, -1], 
                             [1, -1, -1, 1], 
                             [1, 1, 1, 1]])

def Magic():
    if Mode.representation == 'numpy':
        return 1 / np.sqrt(2) * np.array([[1, 1j, 0, 0],
                                          [0, 0, 1j, 1],
                                          [0, 0, 1j, -1],
                                          [1, -1j, 0, 0]])
    else:
        return 1 / sqrt(2) * Matrix([[1, I, 0, 0],
                                     [0, 0, I, 1],
                                     [0, 0, I, -1],
                                     [1, -I, 0, 0]])

def Rxx(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.cos(theta / 2), 0, 0, -1j * np.sin(theta / 2)],
                         [0, np.cos(theta / 2), -1j * np.sin(theta / 2), 0],
                         [0, -1j * np.sin(theta / 2), np.cos(theta / 2), 0],
                         [-1j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)]])
    else:
        return Matrix([[cos(theta / 2), 0, 0, -I * sin(theta / 2)],
                       [0, cos(theta / 2), -I * sin(theta / 2), 0],
                       [0, -I * sin(theta / 2), cos(theta / 2), 0],
                       [-I * sin(theta / 2), 0, 0, cos(theta / 2)]])

def Ryy(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.cos(theta / 2), 0, 0, 1j * np.sin(theta / 2)],
                         [0, np.cos(theta / 2), -1j * np.sin(theta / 2), 0],
                         [0, -1j * np.sin(theta / 2), np.cos(theta / 2), 0],
                         [1j * np.sin(theta / 2), 0, 0, np.cos(theta / 2)]])
    else:
        return Matrix([[cos(theta / 2), 0, 0, I * sin(theta / 2)],
                       [0, cos(theta / 2), -I * sin(theta / 2), 0],
                       [0, -I * sin(theta / 2), cos(theta / 2), 0],
                       [I * sin(theta / 2), 0, 0, cos(theta / 2)]])

def Rzz(theta:float):
    if Mode.representation == 'numpy':
        return np.array([[np.exp(-1j * theta / 2), 0, 0, 0],
                         [0, np.exp(1j * theta / 2), 0, 0],
                         [0, 0, np.exp(1j * theta / 2), 0],
                         [0, 0, 0, np.exp(-1j * theta / 2)]])
    else:
        return Matrix([[exp(-I * theta / 2), 0, 0, 0],
                       [0, exp(I * theta / 2), 0, 0],
                       [0, 0, exp(I * theta / 2), 0],
                       [0, 0, 0, exp(-I * theta / 2)]])

def CAN(tx:float, ty:float, tz:float):
    if Mode.representation == 'numpy':
        return np.array([[np.exp(-1j*tz/2)*np.cos((tx-ty)/2), 0, 0, -1j*np.exp(-1j*tz/2)*np.sin((tx-ty)/2)], 
                         [0, np.exp(1j*tz/2)*np.cos((tx+ty)/2), -1j*np.exp(1j*tz/2)*np.sin((tx+ty)/2), 0], 
                         [0, -1j*np.exp(1j*tz/2)*np.sin((tx+ty)/2), np.exp(1j*tz/2)*np.cos((tx+ty)/2), 0], 
                         [-1j*np.exp(-1j*tz/2)*np.sin((tx-ty)/2), 0, 0, np.exp(-1j*tz/2)*np.cos((tx-ty)/2)]])
    else:
        return Matrix([[exp(-I*tz/2)*cos((tx-ty)/2), 0, 0, -I*exp(-I*tz/2)*sin((tx-ty)/2)], 
                       [0, exp(I*tz/2)*cos((tx+ty)/2), -I*exp(I*tz/2)*sin((tx+ty)/2), 0], 
                       [0, -I*exp(I*tz/2)*sin((tx+ty)/2), exp(I*tz/2)*cos((tx+ty)/2), 0], 
                       [-I*exp(-I*tz/2)*sin((tx-ty)/2), 0, 0, exp(-I*tz/2)*cos((tx-ty)/2)]])


from FrozenYoghourt import *

def scatter(f, lb = 0, ub = 2*np.pi, data_points = 100, point_size = 2, title = '', 
            xlabel = '', ylabel = '', xscale = 'linear', yscale = 'linear', legend_list = []):
    
    """
    Draw scatter plot from the inputted function/s
    
    Parameters
    ----------
    f: function, list
        Individual function or list of functions to plot
    lb: float, optional
        Lower bound of the plot domain (default is 0)
    ub: float, optional
        Upper bound of the plot domain (default is 2pi)
    data_points: int, optional
        Number of scatter points (default is 100)
    point_size: int, optional
        Size of a scatter point (default is 2)
    title: str, optional
        Title of the plot (default is '')
    xlabel: str, optional
        Label of the x-axis (default is '')
    ylabel: str, optional
        Label of the y-axis (default is '')
    xscale: str, optional
        Scale of the x-axis (default is 'linear')
    yscale: str, optional
        Scale of the y-axis (default is 'linear')
    legend_list: list, optional
        List of names for the different plots (default is [])
        
    Returns
    -------
    fig: matplotlib.figure.Figure
        Figure of the plot
    ax: matplotlib.axes._subplots.AxesSubplot
        Axis of the plot
        
    """
    
    domain = np.linspace(lb, ub, data_points)
    fig, ax = plt.subplots()
    ax.set_title(title)
    ax.set_xlabel(xlabel); ax.set_ylabel(ylabel)
    ax.set_xscale(xscale); ax.set_yscale(yscale)

    if (type(f) == list) or (type(f) == tuple):
        codomain_list = [g(domain) for g in f]
        for index, codomain in enumerate(codomain_list):
            ax.scatter(domain, codomain, s = point_size)

        if legend_list == []:
            legend_list = [f'Function {i}' for i in range(1, len(f)+1)]
        ax.legend(legend_list)

    else:
        codomain = f(domain)
        ax.scatter(domain, codomain, s = point_size)
        ax.legend(legend_list)
        
    return fig, ax